try:
    import pyperclip
except ModuleNotFoundError:
    pyperclip=None
import tkinter
try:
    WIN = tkinter.Tk()
    WIN.withdraw()
    CAN=True
except:
    WIN=None
    CAN=False
def copy(text):
    if not CAN:
        raise NotImplementedError("pasting is not implemented on this device")
    if pyperclip:
        return pyperclip.copy(text)
    if WIN:
        return WIN.clipboard_append(text)

    



import string
import warnings
import random
import hashlib
from .error import *
from .letters import *
from .words import words
from .helpers import force_bytes
class Digest:
    def __init__(self,parent,hash_object):
        self.digest=None
        self.hexdigest=None
        self.parent=parent
        self.hash_object=hash_object
        if hasattr(hash_object,'digest'):
            self.digest=hash_object.digest()
        if hasattr(hash_object,'hexdigest'):
            self.hexdigest=hash_object.hexdigest()
    def __repr__(self):
        return f'<Digest for {self.parent} with hash object {self.hash_object}>'

class Hash:
    def __init__(self,hash_object):
        self.hash_object=hash_object
        self.digest=Digest(self,hash_object)
        self.hexdigest=self.digest.hexdigest
        self.bytesdigest=self.digest.digest
    def __repr__(self):
        return f'<Hash with hexdigest {self.digest.hexdigest}>'
class Password:
    def __init__(self,password=None,hashing_method=hashlib.md5,**kwargs):
        self.hashing_method = get_hashing_method(hashing_method)
        self.password=password
        if password is None:
            self.password=randpass(**kwargs).password
        try:
            self.bytes_password=force_bytes(self.password)
        except TypeError:
            raise TypeError("passwords can only be initialized from bytes or str")
    def hash(self,hashing_method=None):
        if hashing_method is None:
            hashing_method = self.hashing_method
        else:
            hashing_method = get_hashing_method(hashing_method)
        return Hash(hashing_method(self.bytes_password))
    def __repr__(self):
        return f'<Password {self.password!r}>'
class Method:
    def __init__(self,m):
        if callable(m):
            self.function=m
        else:
            self.function=METHODS[m]
    def __call__(self,*args,**kwargs):
        return self.function(*args,**kwargs)
def generating_method(fun):
    def decorator(*args,**kwargs):

        number_suffix=kwargs.get('number_suffix',0)
        if 'number_suffix' in kwargs:
            del kwargs['number_suffix']
        hm=kwargs.get('hashing_method',hashlib.md5)
        if 'hashing_method' in kwargs:
            del kwargs['hashing_method']
        number_suffix=Number(number_suffix)
        result=fun(*args,**kwargs)
       
        return Password(result+str(number_suffix),hm)
    return decorator

def _select(min,max=None,step=1):
        if max:
            range_object=range(min,max,step)
        else:
            range_object=range(0,min,step)
        range_object=tuple(range_object)
        return random.choice(range_object)
class Selector:
    def __init__(self,min,max=None,step=1):
        self.min=min
        self.max=max
        self.step=step
    def select(self):
        return _select(self.min,self.max,self.step)
    def __int__(self):
        return self.select()
    def __str__(self):
        return str(self.select())
    
class Number:
    def __init__(self,digitcount=Selector(1,3)):
        self.digitcount=digitcount
    def __repr__(self):
        return repr(self.number)
    @property
    def number(self,final_type=int):
        num=[]
        for q in range(int(self.digitcount)):
            num.append(str(Selector(10)))
        num_string=''.join(num)
        try:
            return final_type(num_string)
        except:
            return num_string
    def __int__(self):
        return self.number
    def __str__(self):
        return str(self.number)
class PasswordGenerator:
    def __init__(self,lowercase=False,uppercase=False,digits=False,symbols=False,all=False,**_):
        self.lowercase=lowercase | all |( not any((uppercase,digits,symbols)))
        self.uppercase=uppercase | all
        self.digits=digits | all
        self.symbols=symbols | all        
        self.choose_from=[]
        if any([self.lowercase,self.uppercase,self.digits,self.symbols]):
            if self.lowercase:
                self.choose_from.extend(string.ascii_lowercase)
            if self.uppercase:
                self.choose_from.extend(string.ascii_uppercase)
            if self.digits:
                self.choose_from.extend(string.digits)
            if self.symbols:
                self.choose_from.extend(string.punctuation)
        else:
            warnings.warn(CategoryWarning('no categories selected. All passwords will be ""'))
        self.passgen=self.generate

    @generating_method
    def generate(self,length=None,min=6,max=20):
        if length is None:
            length=Selector(min,max)
        length=int(length)
        passw=''.join(random.choices(self.choose_from,k=length))
        return passw

    @generating_method
    def pronounceable(self,length=None,min=6,max=20):
        if length is None:
            length=_select(min,max)
        passw=[]
        letters=[consonants,vowels]
        random.shuffle(letters)
        for q in range(length):
            passw.append(random.choice(letters[q%2]))
        return ''.join(passw)
    @generating_method
    def word(self):
        return random.choice(words)
def get_hashing_method(hashing_method):
    if isinstance(hashing_method,str):
        if hashing_method in hashlib.algorithms_guaranteed:
            return getattr(hashlib,hashing_method)
        raise KeyError("Unknown hashing method: {!r}".format(hashing_method))
    if not callable(hashing_method):
        raise TypeError("'hashing_method' must be string or callable")
    return hashing_method
def chars(length=None,min=6,max=20,number_suffix=0,pronounceable=False,**kwargs):
    if pronounceable:
        return _pronounceable(length,min,max,numbezr_suffix,**kwargs)
    return PasswordGenerator(**kwargs).passgen(length,min,max,number_suffix=number_suffix)
def word(number_suffix=0,**kwargs):
    return PasswordGenerator(**kwargs).word(number_suffix=number_suffix)
def _pronounceable(length=None,min=6,max=20,number_suffix=0,**kwargs):
    return PasswordGenerator(**kwargs).pronounceable(length,min,max,number_suffix=number_suffix)
def randpass(method=None,**kwargs):
    
    if method is None:
        method=random.choice([chars,word,_pronounceable])
        if method==word:
            try:
                del kwargs['min']
            except:pass
            try:
                del kwargs['max']
            except:pass
            try:
                del kwargs['length']
            except:pass
    else:
        method=Method(method)
    return method(**kwargs)
def passgen(method=chars,**kwargs):
    if isinstance(method,str):
        if method not in METHODS:
            raise KeyError("Invalid method. Available methods are listed in the docs.")
        method=METHODS[method]
    if not callable(method):
        raise TypeError("{!r} must be a string or callable".format('method'))
    return method(**kwargs)
METHODS={'chars':chars,'characters':chars,'c':chars,'word':word,'w':word,'pronounceable':_pronounceable,'p':_pronounceable}



import click
import hashlib
from click_default_group import DefaultGroup
from . import password
from . import clipboard
from .helpers import get_loop
import sys
import builtins
@click.group(cls=DefaultGroup,default='generate',default_if_no_args=True)
def main():
    return 1

@main.command()
@click.option('--type','-t',default='chars',type=click.Choice(password.METHODS),help='generation method to use')
@click.option('--min',default=6,help='Minimal number of characters in password (has no effect if type is word)')
@click.option('--max',default=20,help='Maximal number of characters in password (has no effect if type is word)')
@click.option('--length','-l',default=None,type=int,help='Precise number of characters in password. Overrides --min and --max (has no effect if type is word)')
@click.option('--number-suffix','-s',default=0,help='Length of number suffix')
@click.option('--copy','-c',is_flag=True,help='Use this to copy the generated password to the clipboard')
@click.option('--hash','-h',is_flag=True,help='Print also the password hash')
@click.option('--algorithm','-a',default='md5',type=click.Choice(hashlib.algorithms_guaranteed),help='Hashing algorithm to use with -h. Default MD5. (If -h is not set, has no effect)')
@click.option('--lowercase','-L',is_flag=True,help='If type is chars, adds lowercase ASCII to the generation set. This flag is automatically set if none of -U,-D,-S,-A,-C is set.')
@click.option('--uppercase','-U',is_flag=True,help='If type is chars, adds uppercase ASCII to the generation set')
@click.option('--digits','-D',is_flag=True,help='If type is chars, adds decimal digits to the generation set')
@click.option('--symbols','-S',is_flag=True,help='If type is chars, adds ASCII symbols to the generation set')
@click.option('--all','-A',is_flag=True,help='If type is chars, adds all groups to the generation set')
@click.option('--copy-hash','-C',is_flag=True,help='Use this to copy the generated hash to clipboard. Has no effect if -h is not set.')
@click.option('--number','-n',default=1,help='Number of passwords to generate')
@click.option('--infinite','-i',is_flag=True,help='Constantly generates passwords. Overrides -n.')
def generate(type,min,max,length,number_suffix,copy,hash,algorithm,lowercase,uppercase,digits,symbols,all,copy_hash,number,infinite):
    kwargs={}
    if (copy or copy_hash) and not clipboard.CAN:
              click.secho("Can't use clipboard - no clipboard detected. Are you on a terminal?",err=True,fg='red')
              return 1
    for _ in get_loop(number,infinite):
        if password.METHODS[type] == password.chars:
            if not all|lowercase|uppercase|digits|symbols:
                lowercase=True
            kwargs.update({'lowercase':lowercase,'uppercase':uppercase,'digits':digits,'symbols':symbols,'all':all})
        passwd = password.passgen(type,min=min,max=max,length=length,number_suffix=number_suffix,**kwargs)
        click.echo(passwd.password)
        if hash:
            h =  passwd.hash(algorithm).hexdigest
            click.echo('[{0} HASH]: {1}'.format(algorithm, h))
            if copy:
                  clipboard.copy(h)
        if copy:
            clipboard.copy(passwd.password)
@main.command()
@click.option('--input','-i',default=sys.stdin,type=click.File('r'),help='Input to read from. Default is stdin')
@click.option('--output','-o',default=sys.stdout.buffer,type=click.File('wb'),help='Output to write to. Default is stdout')
@click.option('--algorithm','-a',default='md5',type=click.Choice(hashlib.algorithms_guaranteed),help='Hashing algorithm to use. Default MD5')
@click.option('--bytes-digest','-b',is_flag=True,help='Output normal digest instead of hexdigest.')
@click.option('--repr','-r',is_flag=True,help='If -b is set, outputs Python internal string representation of bytes output')
def hash(input,output,algorithm,bytes_digest,repr):
    for passwd in input:
        p = password.Password(passwd)
        if not bytes_digest:
            output.write(p.hash(algorithm).hexdigest.encode('ascii'))
            output.write(b'\n')
        else:
            if repr:
                output.write(builtins.repr(p.hash(algorithm).bytesdigest).encode())
                output.write(b'\n')
            else:
                output.write(p.hash(algorithm).bytesdigest)
        output.flush()




def infinite_loop():
    it=0
    while True:
        yield it
        it+=1
def get_loop(n,i):
    if i:
        return infinite_loop()
    return range(n)
def force_bytes(o):
    if isinstance(o,str):
        return o.encode()
    elif isinstance(o,bytes):
        return o
    raise TypeError



import setuptools
setuptools.setup(name="pyopensesame",
                 version="1.0.0",
                 description="Advanced password generator and hasher",
                 long_description="""
# opensesame
Advanced password generator and hasher made in Python.
# Installation
## From PyPI(any OS)
```bash
python3 -m pip install opensesame
```
## Directly from GitHub(Linux only)
```bash
git clone "https://github.com/jenca-adam/opensesame/" opensesame_build
cd opensesame_build
./setup.sh
cd ..
rm -rf opensesame_build
```
You can also use `sudo python3 setup.py install`, but `setup.sh` allows local installation.
# CLI
`opensesame` CLI supports two commands:<br>
1.`generate` for generating passwords(default)

```bash
Usage: python -m opensesame generate [OPTIONS]

Options:
  -t, --type [chars|characters|c|word|w|pronounceable|p]
                                  generation method to use
  --min INTEGER                   Minimal number of characters in password
                                  (has no effect if type is word)
  --max INTEGER                   Maximal number of characters in password
                                  (has no effect if type is word)
  -l, --length INTEGER            Precise number of characters in password.
                                  Overrides --min and --max (has no effect if
                                  type is word)
  -s, --number-suffix INTEGER     Length of number suffix
  -c, --copy                      Use this to copy the generated password to
                                  the clipboard
  -h, --hash                      Print also the password hash
  -a, --algorithm [sha3_384|sha512|blake2b|sha1|sha224|sha3_512|sha256|shake_256|md5|sha384|shake_128|sha3_256|sha3_224|blake2s]
                                  Hashing algorithm to use with -h. Default
                                  MD5. (If -h is not set, has no effect)
  -L, --lowercase                 If type is chars, adds lowercase ASCII to
                                  the generation set. This flag is
                                  automatically set if none of -U,-D,-S,-A,-C
                                  is set.
  -U, --uppercase                 If type is chars, adds uppercase ASCII to
                                  the generation set
  -D, --digits                    If type is chars, adds decimal digits to the
                                  generation set
  -S, --symbols                   If type is chars, adds ASCII symbols to the
                                  generation set
  -A, --all                       If type is chars, adds all groups to the
                                  generation set
  -C, --copy-hash                 Use this to copy the generated hash to
                                  clipboard. Has no effect if -h is not set.
  -n, --number INTEGER            Number of passwords to generate
  -i, --infinite                  Constantly generates passwords. Overrides
                                  -n.
  --help                          Show this message and exit.
```

2.`hash` for hashing passwords.<br>
```bash
Usage: python -m opensesame hash [OPTIONS]

Options:
  -i, --input FILENAME            Input to read from. Default is stdin
  -o, --output FILENAME           Output to write to. Default is stdout
  -a, --algorithm [sha224|blake2b|sha3_384|sha384|blake2s|sha1|shake_128|sha3_224|sha512|sha3_256|sha3_512|sha256|md5|shake_256]
                                  Hashing algorithm to use. Default MD5
  -b, --bytes-digest              Output normal digest instead of hexdigest.
  -r, --repr                      If -b is set, outputs Python internal string
                                  representation of bytes output
  --help                          Show this message and exit.
```
# Python lib
## Generation methods
# opensesame
Advanced password generator and hasher made in Python.
# Installation
## From PyPI(any OS)
```bash
python3 -m pip install opensesame
```
## Directly from GitHub(Linux only)
```bash
git clone "https://github.com/jenca-adam/opensesame/" opensesame_build
cd opensesame_build
./setup.sh
cd ..
rm -rf opensesame_build
```
You can also use `sudo python3 setup.py install`, but `setup.sh` allows local installation.
# CLI
`opensesame` CLI supports two commands:<br>
1.`generate` for generating passwords(default)

```bash
Usage: python -m opensesame generate [OPTIONS]

Options:
  -t, --type [chars|characters|c|word|w|pronounceable|p]
                                  generation method to use
  --min INTEGER                   Minimal number of characters in password
                                  (has no effect if type is word)
  --max INTEGER                   Maximal number of characters in password
                                  (has no effect if type is word)
  -l, --length INTEGER            Precise number of characters in password.
                                  Overrides --min and --max (has no effect if
                                  type is word)
  -s, --number-suffix INTEGER     Length of number suffix
  -c, --copy                      Use this to copy the generated password to
                                  the clipboard
  -h, --hash                      Print also the password hash
  -a, --algorithm [sha3_384|sha512|blake2b|sha1|sha224|sha3_512|sha256|shake_256|md5|sha384|shake_128|sha3_256|sha3_224|blake2s]
                                  Hashing algorithm to use with -h. Default
                                  MD5. (If -h is not set, has no effect)
  -L, --lowercase                 If type is chars, adds lowercase ASCII to
                                  the generation set. This flag is
                                  automatically set if none of -U,-D,-S,-A,-C
                                  is set.
  -U, --uppercase                 If type is chars, adds uppercase ASCII to
                                  the generation set
  -D, --digits                    If type is chars, adds decimal digits to the
                                  generation set
  -S, --symbols                   If type is chars, adds ASCII symbols to the
                                  generation set
  -A, --all                       If type is chars, adds all groups to the
                                  generation set
  -C, --copy-hash                 Use this to copy the generated hash to
                                  clipboard. Has no effect if -h is not set.
  -n, --number INTEGER            Number of passwords to generate
  -i, --infinite                  Constantly generates passwords. Overrides
                                  -n.
  --help                          Show this message and exit.
```

2.`hash` for hashing passwords.<br>
```bash
Usage: python -m opensesame hash [OPTIONS]

Options:
  -i, --input FILENAME            Input to read from. Default is stdin
  -o, --output FILENAME           Output to write to. Default is stdout
  -a, --algorithm [sha224|blake2b|sha3_384|sha384|blake2s|sha1|shake_128|sha3_224|sha512|sha3_256|sha3_512|sha256|md5|shake_256]
                                  Hashing algorithm to use. Default MD5
  -b, --bytes-digest              Output normal digest instead of hexdigest.
  -r, --repr                      If -b is set, outputs Python internal string
                                  representation of bytes output
  --help                          Show this message and exit.
```
# Python lib
## Generation methods
### chars
This generates a password of a given length randomly selecting from the given character sets. Character sets are: 

* `lowercase` -> lower case ascii chars
* `uppercase` -> UPPER CASE ASCII CHARS
* `digits` -> 0123456789
* `symbols` -> *!"#$%&\'()+,\*-./:;<=>?@[\\]^_`{|}~*
* `all` -> everything above
Character sets are added by setting kwarg `<CHARSET_NAME>` to `True`.
If no character sets are specified, only lowercase is used.

You can specify length using `length`,`min` and `max` atributtes, where `length` overrides the boundaries(`min` and `max`).
Default length is picked between 6 and 20.
> :warning: Be careful! Don't select only one boundary, as it might cause `IndexError`.

You can also generate a pronounciable password by setting `pronounciable` to `True`.
Character sets don't apply to pronounciable passwords.
### word
This just selects a random word from the wordlist.
No attributes from `chars` apply to this.
### passgen
With `passgen`,you can select a method and call the method with given `**kwargs`. <br>
It's just a wrapper. <br>
The methods are selected by setting the first arg of `passgen` to `chars`,`word` or `pronounciable`. You can also set it to the first char of their name. 
## `number_suffix`
`chars`,`word` and `passgen` all have the argument `number_suffix`. It specifies how many digits are supposed to be added to the end of the password.
## `Password` object
Generation methods return this.<br>
It's a string wrapped in hashing.
Hashing is done with  `Password.hash(algorithm)`.
`Password.hash` returns `Hash` object with attributes `hexdigest`(hexadecimal representation of the digest) and `bytesdigest`(actual digest)


### chars
This generates a password of a given length randomly selecting from the given character sets. Character sets are: 

* `lowercase` -> lower case ascii chars
* `uppercase` -> UPPER CASE ASCII CHARS
* `digits` -> 0123456789
* `symbols` -> *!"#$%&\'()+,\*-./:;<=>?@[\\]^_`{|}~*
* `all` -> everything above
Character sets are added by setting kwarg `<CHARSET_NAME>` to `True`.
If no character sets are specified, only lowercase is used.

You can specify length using `length`,`min` and `max` atributtes, where `length` overrides the boundaries(`min` and `max`).
Default length is picked between 6 and 20.
> :warning: Be careful! Don't select only one boundary, as it might cause `IndexError`.

You can also generate a pronounciable password by setting `pronounciable` to `True`.
Character sets don't apply to pronounciable passwords.
### word
This just selects a random word from the wordlist.
No attributes from `chars` apply to this.
### passgen
With `passgen`,you can select a method and call the method with given `**kwargs`. <br>
It's just a wrapper. <br>
The methods are selected by setting the first arg of `passgen` to `chars`,`word` or `pronounciable`. You can also set it to the first char of their name. 
## `number_suffix`
`chars`,`word` and `passgen` all have the argument `number_suffix`. It specifies how many digits are supposed to be added to the end of the password.
## `Password` object
Generation methods return this.<br>
It's a string wrapped in hashing.
Hashing is done with  `Password.hash(algorithm)`.
`Password.hash` returns `Hash` object with attributes `hexdigest`(hexadecimal representation of the digest) and `bytesdigest`(actual digest)
""",
    long_description_content_type='text/markdown',
    packages=['opensesame'],
    package_data={"opensesame":["list.txt"]},
    include_package_data=True,
    entry_points={"console_scripts":["opensesame = opensesame.main_click:main"]},
    author="Adam Jenca",
    classifiers=[ "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
                 "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                 ],
    author_email="jenca.a@gjh.sk",
    project_urls={
            "GitHub":"https://github.com/jenca-adam/opensesame/"
            }
)

